(************************************************************************
 Author:	Robert Rossmair
		Wagelwitz, Rodaer Str. 6
		04688 Mutzschen
		Germany

	e-mail:	Robert.Rossmair@t-online.de

 Module:	MDIWallp

 Version:	4.0, 02-Jun-1999

 History:
   02-Jun-99: Version 4.0
        New properties:
          GrayMapped: inspired by a suggestion by Harald Marschalek;
                      not available under Delphi 2
          AutoSizeTile
          TileWidth
          TileHeight
        Using transparent bitmaps is less time consuming now; they
        are drawn on an offscreen bitmap, which is used instead.
        General code review
   02-Feb-99: Version 3.1
        Fixed bug which prevented background from being updated when
        Picture property is set to nil.
        Added Enabled property
   13-Aug-98: Version 3.0.1
        As a goodie for D2 users, ConnectToClient sets the
        proper (WS_EX_CLIENTEDGE) border style for the MDI
        client window.
   06-Aug-98: Version 3.0
        introduces Picture, Shift, ShiftMode and Info properties.
        Some code sent to me by Kristoffer Henriksson pointed me
        to a way how to overcome the problems which caused me to
        remove the Mode property in V2.0. This allowed me to re-
        implement the Mode property again.
   25-Jun-98: Version 2.1.1
        Cosmetic changes
   11-Jun-98: Version 2.1.0
        Zigzag property implemented (using code by Fabio Cavicchio)
   26-Apr-98: Version 2.0.0
        This version fixes the bug that version 1.6 only claimed to
        fix...(versions < 2.0.0 will not redraw correctly if a MDI
        child window which has been moved partially out of view is
        scrolled in view again (unfortunately this hadn't been tested
        ever).
        Note that I found the values for internal constants ScrollUnit
        and ScrollLineSize by experimenting, so there is no guarantee
        that they are valid under all circumstances. However, tests
        did not give any hints that they are not.
        The Mode property has been removed. The virtual desktop of
        a MDI frame window is best considered as infinite in size
        (you may move a MDI child to the left until it's almost
        completely outside the window, then scroll it in view, then
        again move it to the left and so forth...); therefore notions
        like 'top', 'left', 'center' are meaningless in this context.
        The component now always works in a mode equivalent to
        Mode=wpTile in earlier versions.
   21-Apr-98: Version 1.6
   06-Mar-98: Version 1.5
	Added transparency code by Mike Burton (see below),
	with some modifications, among others:
	- SetTransparent property access method
	- TransparentColor property
	- TransparentMode property
   #########################################
   05-Mar-1998
      Michael Burton AMS-Services, Inc.
      burtonmi@ams.iix.com

      Add a transparent background to the
      bitmap. Use the upper-left pixel of
      the bitmap as the transparent color.

      Add a Transparent property to control
      the transparent background.
   #########################################
   03-Mar-98: Version 1.4
	Bug fix: Prior versions will not redraw correctly on
	form resize if Mode property not in [wpTile, wpTopLeft].
   09-Dez-97: Version 1.3
	Make it work with non-MDI windows by commenting
	out two lines in TMDIWallpaper.Create.
	Minor fix.
   28-Nov-97: Version 1.2
	Added wpTopRight, wpBottomLeft, wpBottomRight.
	Some fixes.
   06-Nov-97: Version 1.1
	added wpStretch
   22-Oct-97: released Version 1.0

 Description:	Background bitmap component for MDI frame windows

 Copyright  1997-1999  Robert Rossmair

 Use, modification and distribution is allowed without limitation,
 warranty, or liability of any kind.

 If you use this software or portions of it in a commercial program,
 I would appreciate it if you mentioned my contribution in the
 documentation materials accompaning your work.
************************************************************************)
unit MDIWallp;

interface
{ ***************** Conditional Compiler Symbols **********************

  RECOGNIZE_GIF        use GIFImage library by Anders Melander et alii
                       (download address:
                       http://www.melander.dk/delphi/gifimage/index.html)

  NO_JPEG              $DEFINE this if you don't have Borland's JPEG
                       library
  ********************************************************************* }

{_$DEFINE RECOGNIZE_GIF}
{_$DEFINE NO_JPEG}

uses SysUtils, Windows, Messages, Classes, Graphics,
{$IFNDEF VER90}
  rrColors,
  {$IFNDEF NO_JPEG}
  JPEG,
  {$ENDIF}
{$ENDIF}
{$IFDEF RECOGNIZE_GIF}
  GIFImage,
{$ENDIF}
  Controls, Forms, CompInfo, CopyrInf;

type
  TWallpaperMode = (wpTile, wpCenter, wpTopLeft, wpStretch, wpTopRight, wpBottomLeft, wpBottomRight);
{$IFDEF VER90}
  TTransparentMode = (tmAuto, tmFixed);
{$ENDIF}
  EMDIWallpaper = class(Exception);
  TShiftMode = (smRows, smColumns);

  TMDIWallpaper = class(TComponent)
  private
    FClientInstance: Pointer;
    FDefClientProc: TFarProc;
    FMDIFormInstance: Pointer;
    FDefMDIFormProc: TFarProc;
    FPicture: TPicture;
    FMode: TWallpaperMode;
    FEnabled: Boolean;
    FTransparent: Boolean;
    FTransparentMode: TTransparentMode;
    FTransparentColor: TColor;
    FInfo: TComponentInfo;
    FHorzOffset: Integer;
    FVertOffset: Integer;
    FHOffset: Integer;
    FVOffset: Integer;
    FTileWidth: Integer;
    FTileHeight: Integer;
    FWorkingBmp: TBitmap;
    FShift: Cardinal;
    FShiftMode: TShiftMode;
    FZigzag: Boolean;
    FAutoSizeTile: Boolean;
    FInUpdWorkingBmp: Boolean;
    {$IFNDEF VER90}
    FGrayMapped: Boolean;
    procedure MDIFormWndProc(var Message: TMessage);
    procedure SetGrayMapped(Value: Boolean);
    procedure SysColorChange;
    {$ENDIF}
    procedure ClientInvalidate;
    procedure ClientWndProc(var Message: TMessage);
    procedure ConnectToClient;
    function FirstVisibleRow: Integer;
    function GetClient: TForm;
    function GetClientHandle: hWnd;
    function GetTransparentColor: TColor;
    procedure PictureChanged(Sender: TObject);
    procedure SetAutoSizeTile(Value: Boolean);
    procedure SetEnabled(Value: Boolean);
    procedure SetInfo(Value: TComponentInfo);
    procedure SetMode(Value: TWallpaperMode);
    procedure SetPicture(Value: TPicture);
    procedure SetShift(Value: Cardinal);
    procedure SetShiftMode(Value: TShiftMode);
    procedure SetTileWidth(Value: Integer);
    procedure SetTileHeight(Value: Integer);
    procedure SetTransparent(Value: Boolean);
    procedure SetTransparentColor(Value: TColor);
    procedure SetTransparentMode(Value: TTransparentMode);
    procedure SetZigZag(Value: Boolean);
    procedure TileGraphic(Graphic: TGraphic; Canvas: TCanvas; Left, Top, Width, Height: Integer);
    function TransparentColorStored: Boolean;
    procedure UpdateHOffset;
    procedure UpdateVOffset;
    procedure UpdateWorkingBmp;
  protected
    destructor Destroy; override;
  public
    constructor Create(AOwner: TComponent); override;
    property Client: TForm read GetClient;
    property ClientHandle: hWnd read GetClientHandle;
  published
    property AutoSizeTile: Boolean read FAutoSizeTile write SetAutoSizeTile;
    property Enabled: Boolean read FEnabled write SetEnabled default True;
    {$IFNDEF VER90}
    property GrayMapped: Boolean read FGrayMapped write SetGrayMapped;
    {$ENDIF}
    property Info: TComponentInfo read FInfo write SetInfo;
    property Mode: TWallpaperMode read FMode write SetMode;
    property Picture: TPicture read FPicture write SetPicture;
    property TileWidth: Integer read FTileWidth write SetTileWidth;
    property TileHeight: Integer read FTileHeight write SetTileHeight;
    property Transparent: Boolean read FTransparent write SetTransparent;
    property TransparentColor: TColor read GetTransparentColor
      write SetTransparentColor stored TransparentColorStored;
    property TransparentMode: TTransparentMode read FTransparentMode
      write SetTransparentMode default tmAuto;
    property Shift: Cardinal read FShift write SetShift;
    property ShiftMode: TShiftMode read FShiftMode write SetShiftMode;
    property Zigzag: Boolean read FZigzag write SetZigzag;
  end;

procedure Register;

implementation

uses DsgnIntf, PicEdit;

const
  ScrollLineSize = 3;
  ScrollUnit     = 8;
{$IFNDEF VER90}
type
  TColorGradation = array[Byte] of TColor;
  PColorGradation = ^TColorGradation;
var
  FSysColorGradation: PColorGradation;

procedure UpdateSysColorGradation;
var
  SysHLS: THLSVector;
  FaceLum, MaxLum: THLSValue;
  i: Integer;
begin
  if FSysColorGradation = nil then New(FSysColorGradation);
  SysHLS := RGBtoHLS(ColorToRGB(clBtnHighlight));
  MaxLum := SysHLS.Luminance;
  SysHLS := RGBtoHLS(ColorToRGB(clBtnFace));
  FaceLum := SysHLS.Luminance;
  with SysHLS do
  begin
    for i := 0 to 192 do
    begin
      Luminance := i * FaceLum div 192;
      FSysColorGradation[i] := HLStoRGB(Hue, Luminance, Saturation);
    end;
    for i := 193 to 255 do
    begin
      Luminance := FaceLum + (MaxLum-FaceLum) * (i-192) div (255-192);
      FSysColorGradation[i] := HLStoRGB(Hue, Luminance, Saturation);
    end;
  end;
end;
{$ENDIF}
procedure Error(Text: string);
var
  StackTop: record end;
  Stack: record
    EBP: Integer;
    ReturnAddress: Pointer;
  end absolute StackTop;
begin
  raise EMDIWallpaper.Create(Text) at Stack.ReturnAddress;
end;

constructor TMDIWallpaper.Create(AOwner: TComponent);
var
  i: Integer;
begin
  if not (AOwner is TForm) then
    Error('TMDIWallpaper.Owner needs to be a TForm')
{ commented out to make this work with non-MDI windows, too:
   else if TForm(AOwner).FormStyle <> fsMDIForm then
    Error('This is no MDI frame window')
}
  else with AOwner do for i := 0 to ComponentCount-1 do if Components[i] is TMDIWallpaper then
    Error(Format('%s: Only one instance of TMDIWallpaper allowed', [AOwner.Name]));
  inherited Create(AOwner);
  FTransparentColor := clDefault;
  FPicture := TPicture.Create;
  FPicture.OnChange := PictureChanged;
  TForm(Owner).HandleNeeded;
  FEnabled := True;
  ConnectToClient;
  FInfo := TComponentInfo.Create(Self,
           CopyrightOwner,
           '1997-1999',
           '4.0',
           FreewareDisclaimer+#13#13'Acknowledgements (for ideas and code):'#13+
           '    Michael Burton AMS-Services, Inc.'#13+
           '    Fabio Cavicchio'#13+
           '    Kristoffer Henriksson'#13+
           '    Harald Marschalek',
           WebsiteURL,
           MailAddr);
end;

destructor TMDIWallpaper.Destroy;

  procedure ChainError;
  begin
    Error(Format('%s: Oops...Messing up %s''s window procedure chain', [Name, Client.Name]));
  end;
  
begin
  FInfo.Free;
  if Assigned(Owner) and not (csDesigning in ComponentState) then
  begin
    if Client.HandleAllocated then
    begin
{ Falls jemand Anderer es wagt, in dreister Miachtung der
  Maximen objektorientierter Programmierung ebenfalls in
  den Innereien des Clients herumzupfuschen (WIR drfen das),
  ist hier Gelegenheit, das ans Licht zu bringen:
FREE translation:
  If anybody else dares to tamper with the Client (WE are allowed to do
  so-), we'll find it out NOW:
}
      if (SetWindowLong(ClientHandle, GWL_WNDPROC, Longint(FDefClientProc)) <> Longint(FClientInstance))
      and not (csDestroying in Client.ComponentState) then
        ChainError;
      FreeObjectInstance(FClientInstance);
      if Assigned(FMDIFormInstance) then
      begin
        if (SetWindowLong(Client.Handle, GWL_WNDPROC, Longint(FDefMDIFormProc)) <> Longint(FMDIFormInstance))
        and not (csDestroying in Client.ComponentState) then
          ChainError;
        FreeObjectInstance(FMDIFormInstance);
      end;
    end
    else FreeObjectInstance(FDefClientProc);
  end;
  FPicture.Free;
  FWorkingBmp.Free;
  inherited;
end;

procedure TMDIWallpaper.ConnectToClient;
{ This is the dirty way, but heck, who knows a better one? }
begin
  if csDesigning in ComponentState then Exit;
  FClientInstance := MakeObjectInstance(ClientWndProc);
  FDefClientProc := Pointer(SetWindowLong(ClientHandle, GWL_WNDPROC, Longint(FClientInstance)));
  if Client.FormStyle = fsMDIForm then
  begin
{$IFDEF VER90}
  // a little goodie...
    SetWindowLong(ClientHandle, GWL_EXSTYLE, WS_EX_CLIENTEDGE);
{$ELSE}    
    FMDIFormInstance := MakeObjectInstance(MDIFormWndProc);
    FDefMDIFormProc := Pointer(SetWindowLong(Client.Handle, GWL_WNDPROC, Longint(FMDIFormInstance)));
{$ENDIF}
  end;
end;

procedure TMDIWallpaper.TileGraphic(Graphic: TGraphic; Canvas: TCanvas; Left, Top, Width, Height: Integer);
var
  i, j: Integer;
  iMin: Integer;
  S, OddShift: Integer;
begin
  OddShift := 0;  // just to satisfy the compiler
  if FShiftMode = smRows then
  begin
    if FZigZag then
    begin
      OddShift := FTileWidth div 2;
      if Odd(FirstVisibleRow)
        then S := OddShift
        else S := 0;
    end else
    begin
      S := (FirstVisibleRow * FShift) mod FTileWidth;
      if S < 0 then Inc(S, FTileWidth);
    end;
    for j := 0 to (Height-1) div FTileHeight do
    begin
      if S = 0 then iMin := 0 else iMin := -1;
      for i := iMin to (Width-1) div FTileWidth do
        Canvas.Draw(Left+i*FTileWidth+S, Top+j*FTileHeight, Graphic);
      if FZigZag then
        S := S xor OddShift
      else
      begin
        Inc(S, FShift);
        S := S mod FTileWidth;
      end;
    end;
  end else
  begin
    if FZigZag then
    begin
      OddShift := FTileHeight div 2;
      if Odd(FirstVisibleRow)
        then S := OddShift
        else S := 0;
    end else
    begin
      S := (FirstVisibleRow * FShift) mod FTileHeight;
      if S < 0 then Inc(S, FTileHeight);
    end;
    for i := 0 to (Width-1) div FTileWidth do
    begin
      if S = 0 then iMin := 0 else iMin := -1;
      for j := iMin to (Height-1) div FTileHeight do
        Canvas.Draw(Left+i*FTileWidth, Top+j*FTileHeight+S, Graphic);
      if FZigZag then
        S := S xor OddShift
      else
      begin
        Inc(S, FShift);
        S := S mod FTileHeight;
      end;
    end;
  end;
end;

procedure TMDIWallpaper.ClientInvalidate;
begin
  if ComponentState * [csReading,  csDesigning] = [] then
    InvalidateRect(ClientHandle, nil, True);
end;

procedure TMDIWallpaper.ClientWndProc(var Message: TMessage);
var
  ClientIsMDIForm: Boolean;

  procedure PaintBackground(Graphic: TGraphic);
  var
    R: TRect;
    X, Y, W, H: Integer;
  begin
    with Client.Canvas do
    begin
      Handle := TWMEraseBkGnd(Message).DC;
      Windows.GetClientRect(ClientHandle, R);
      W := R.Right-R.Left;
      H := R.Bottom-R.Top;
      if FMode = wpTile then
	TileGraphic(Graphic, Client.Canvas,
          FHOffset,
          FVOffset,
          W-FHOffset,
          H-FVOffset)
      else
      begin
        if ClientIsMDIForm then
        begin
        // We don't want the background to move
        // when scrollbars appear or disappear:
          if (GetWindowLong(ClientHandle, GWL_STYLE) and WS_HSCROLL) <> 0 then
            Inc(H, GetSystemMetrics(SM_CYHSCROLL));
          if (GetWindowLong(ClientHandle, GWL_STYLE) and WS_VSCROLL) <> 0 then
            Inc(W, GetSystemMetrics(SM_CXVSCROLL));
        end else
        begin
          R.Left := FHorzOffset;
          R.Top := FVertOffset;
          with Client.HorzScrollBar do if Range > W then W := Range;
          with Client.VertScrollBar do if Range > H then H := Range;
        end;
        R.Right := R.Left + W;
        R.Bottom := R.Top + H;
        Brush := Client.Brush;
        case FMode of
          wpCenter:
                  begin
                    X := R.Left+(W - Graphic.Width) div 2;
                    Y := R.Top+(H - Graphic.Height) div 2;
                    FillRect(Rect(R.Left, R.Top, R.Right, Y));
                    FillRect(Rect(R.Left, Y, X, Y+Graphic.Height));
                    FillRect(Rect(X+Graphic.Width, Y, R.Right, Y+Graphic.Height));
                    FillRect(Rect(R.Left, Y+Graphic.Height, R.Right, R.Bottom));
                    Draw(X, Y, Graphic);
                  end;
          wpStretch:
                  begin
                    StretchDraw(R, Graphic);
                  end;
          wpTopLeft:
                  begin
                    FillRect(Rect(R.Left+Graphic.Width, R.Top, R.Right, R.Top+Graphic.Height));
                    FillRect(Rect(R.Left, R.Top+Graphic.Height, R.Right, R.Bottom));
                    Draw(R.Left, R.Top, Graphic);
                  end;
          wpTopRight:
                  begin
                    FillRect(Rect(R.Left, R.Top, R.Right-Graphic.Width, R.Top+Graphic.Height));
                    FillRect(Rect(R.Left, R.Top+Graphic.Height, R.Right, R.Bottom));
                    Draw(R.Right-Graphic.Width, R.Top, Graphic);
                  end;
          wpBottomLeft:
                  begin
                    FillRect(Rect(R.Left, R.Top, R.Right, R.Bottom-Graphic.Height));
                    FillRect(Rect(R.Left+Graphic.Width, R.Bottom-Graphic.Height, R.Right, R.Bottom));
                    Draw(R.Left, R.Bottom-Graphic.Height, Graphic);
                  end;
          wpBottomRight:
                  begin
                    FillRect(Rect(R.Left, R.Top, R.Right, R.Bottom-Graphic.Height));
                    FillRect(Rect(R.Left, R.Bottom-Graphic.Height, R.Right-Graphic.Width, R.Bottom));
                    Draw(R.Right-Graphic.Width, R.Bottom-Graphic.Height, Graphic);
                  end;
        end;
      end;
      Handle := 0;
    end;
  end;

  function EraseBkgnd: Boolean;
  var
    Graphic: TGraphic;
  begin
    Result := False;
    if (FPicture.Width > 0) and (FPicture.Height > 0) then
    begin
      Graphic := FWorkingBmp;
      if Graphic = nil then
        Graphic := FPicture.Graphic;
      PaintBackground(Graphic);
      Result := True;
    end;
  end;

  procedure CalcHorzOffset;
  var
    ScrollInfo: TScrollInfo;
    dx, MaxChange: Integer;
  begin
    if ClientIsMDIForm then
    begin
      ScrollInfo.cbSize := SizeOf(ScrollInfo);
      ScrollInfo.fMask := SIF_ALL;
      GetScrollInfo(ClientHandle, SB_HORZ, ScrollInfo);
      dx := 0;
      case TWMHScroll(Message).ScrollCode of
        SB_LINELEFT:
        begin
          dx := ScrollInfo.nPos-ScrollInfo.nMin;
          if dx > ScrollLineSize then dx := ScrollLineSize;
        end;
        SB_LINERIGHT:
        with ScrollInfo do
        begin
          dx := nPage-1;
          if dx < 0 then dx := 0;
          dx := nPos-(nMax-dx);
          if dx < -ScrollLineSize then dx := -ScrollLineSize;
        end;
        SB_PAGELEFT:
        with ScrollInfo do
        begin
          dx := nPage-1;
          if dx < 0 then dx := 0;
          if dx > nPos-nMin then dx := nPos-nMin;
        end;
        SB_PAGERIGHT:
        with ScrollInfo do
        begin
          dx := nPage-1;
          if dx < 0 then dx := 0;
          MaxChange := (nMax-dx)-nPos;
          if dx > MaxChange then dx := MaxChange;
          dx := -dx;
        end;
        SB_THUMBPOSITION: dx := -TWMHScroll(Message).Pos;
      end;
      Inc(FHorzOffset, dx * ScrollUnit);
    end
    else FHorzOffset := -Client.HorzScrollBar.Position;
    UpdateHOffset;
  end;

  procedure CalcVertOffset;
  var
    ScrollInfo: TScrollInfo;
    dy, MaxChange: Integer;
  begin
    if ClientIsMDIForm then
    begin
      ScrollInfo.cbSize := SizeOf(ScrollInfo);
      ScrollInfo.fMask := SIF_ALL;
      GetScrollInfo(ClientHandle, SB_VERT, ScrollInfo);
      dy := 0;
      case TWMVScroll(Message).ScrollCode of
        SB_LINELEFT:
        begin
          dy := ScrollInfo.nPos-ScrollInfo.nMin;
          if dy > ScrollLineSize then dy := ScrollLineSize;
        end;
        SB_LINERIGHT:
        with ScrollInfo do
        begin
          dy := nPage-1;
          if dy < 0 then dy := 0;
          dy := nPos-(nMax-dy);
          if dy < -ScrollLineSize then dy := -ScrollLineSize;
        end;
        SB_PAGELEFT:
        with ScrollInfo do
        begin
          dy := nPage-1;
          if dy < 0 then dy := 0;
          if dy > nPos-nMin then dy := nPos-nMin;
        end;
        SB_PAGERIGHT:
        with ScrollInfo do
        begin
          dy := nPage-1;
          if dy < 0 then dy := 0;
          MaxChange := (nMax-dy)-nPos;
          if dy > MaxChange then dy := MaxChange;
          dy := -dy;
        end;
        SB_THUMBPOSITION: dy := -TWMVScroll(Message).Pos;
      end;
      Inc(FVertOffset, dy * ScrollUnit);
    end
    else FVertOffset := -Client.VertScrollBar.Position;
    UpdateVOffset;
  end;

  procedure InvalidateBackground;
  begin
    InvalidateRect(ClientHandle, nil, True);
  end;

begin
  ClientIsMDIForm := Client.FormStyle = fsMDIForm;
  with Message do
  begin
    if Msg = WM_ERASEBKGND then
    if FEnabled and EraseBkgnd then
    begin
      Result := 1;
      Exit;
    end;
    if not ClientIsMDIForm then
      Result := CallWindowProc(FDefClientProc, ClientHandle, Msg, wParam, lParam);
    case Msg of
      WM_SIZE:
        begin
          if not ClientIsMDIForm then
          begin
            CalcHorzOffset;
            CalcVertOffset;
          end;
          if not (FMode in [wpTile, wpTopLeft]) then
            InvalidateBackground;
        end;
    {$IFNDEF VER90}
      WM_SYSCOLORCHANGE: SysColorChange;
    {$ENDIF}
      WM_HSCROLL:
        begin
          CalcHorzOffset;
          if FMode <> wpTile then InvalidateBackground;
        end;
      WM_VSCROLL:
        begin
          CalcVertOffset;
          if FMode <> wpTile then InvalidateBackground;
        end;
    end;
    if ClientIsMDIForm then
      Result := CallWindowProc(FDefClientProc, ClientHandle, Msg, wParam, lParam);
  end;
end;

function TMDIWallpaper.FirstVisibleRow: Integer;
begin
  if FShiftMode = smRows then
  begin
    Result := -FVertOffset div FTileHeight;
    if FVertOffset > 0 then Dec(Result);
  end else
  begin
    Result := -FHorzOffset div FTileWidth;
    if FHorzOffset > 0 then Dec(Result);
  end;
end;

function TMDIWallpaper.GetClient: TForm;
begin
  Result := TForm(Owner);
end;

function TMDIWallpaper.GetClientHandle: hWnd;
begin
  with Client do
  if ClientHandle <> 0
    then Result := ClientHandle
    else Result := Handle;
end;

function TMDIWallpaper.GetTransparentColor: TColor;
var
  Bmp: TBitmap;
begin
  Bmp := nil;
  if FTransparentColor = clDefault then
  if FPicture.Graphic is TBitmap then
    Bmp := TBitmap(FPicture.Graphic)
  {$IFDEF RECOGNIZE_GIF}
  else if FPicture.Graphic is TGIFImage then
    Bmp := TGIFImage(FPicture.Graphic).Bitmap
  {$ENDIF};
  if Assigned(Bmp) then
  begin
    if Bmp.Monochrome then
      Result := clWhite
    else
      Result := Bmp.Canvas.Pixels[0, Bmp.Height - 1];
  end
  else Result := ColorToRGB(FTransparentColor);
  Result := Result or $02000000;
end;

procedure TMDIWallPaper.PictureChanged(Sender: TObject);
begin
  UpdateHOffset;
  UpdateVOffset;
  if (FTileWidth < Picture.Width)
  or (FTileHeight < Picture.Height)
  or (AutoSizeTile and (Picture.Width > 0) and (Picture.Height > 0)) then
  begin
    FTileWidth := Picture.Width;
    FTileHeight := Picture.Height;
  end;
  {$IFNDEF VER90}
  with Picture do
  begin
  if Graphic <> nil then
      Graphic.Transparent := FTransparent;
  end;
  {$ENDIF}
  UpdateWorkingBmp;
end;

procedure TMDIWallpaper.SetAutoSizeTile(Value: Boolean);
begin
  if FAutoSizeTile <> Value then
  begin
    FAutoSizeTile := Value;
    if Mode = wpTile then
      if (TileWidth <> Picture.Width)
      or (TileHeight <> Picture.Height) then
        PictureChanged(Self);
  end;
end;

procedure TMDIWallpaper.SetEnabled(Value: Boolean);
begin
  if Value <> FEnabled then
  begin
    FEnabled := Value;
    ClientInvalidate;
  end;
end;
{$IFNDEF VER90}
procedure TMDIWallpaper.MDIFormWndProc(var Message: TMessage);
begin
  with Message do
  begin
    Result := CallWindowProc(FDefMDIFormProc, Client.Handle, Msg, wParam, lParam);
    if Msg = WM_SYSCOLORCHANGE then SysColorChange;
  end;
end;

procedure TMDIWallpaper.SysColorChange;
begin
  if FGrayMapped then
  begin
    if FPicture.Graphic <> nil then UpdateSysColorGradation;
    UpdateWorkingBmp;
  end;
end;

procedure TMDIWallpaper.SetGrayMapped(Value: Boolean);
begin
  if Value <> FGrayMapped then
  begin
    if Value = True then
      UpdateSysColorGradation;
    FGrayMapped := Value;
    UpdateWorkingBmp;
  end;
end;
{$ENDIF}
procedure TMDIWallpaper.SetInfo(Value: TComponentInfo);
begin
// dummy method to make Info property visible in Object Inspector
end;

procedure TMDIWallpaper.SetMode(Value: TWallpaperMode);
var
  TileModeChanged: Boolean;
begin
  if Value <> FMode then
  begin
    TileModeChanged := (FMode = wpTile) or (Value = wpTile);
    FMode := Value;
    if TileModeChanged
    and ((FTileWidth <> Picture.Width) or (FTileHeight <> Picture.Height))
      then PictureChanged(Self)
      else ClientInvalidate;
  end;
end;

procedure TMDIWallPaper.SetPicture(Value: TPicture);
begin
  FPicture.Assign(Value);
end;

procedure TMDIWallpaper.SetShift(Value: Cardinal);
begin
  if Value <> FShift then
  begin
    FShift := Value;
    FZigZag := False;
    if FMode = wpTile then ClientInvalidate;
  end;
end;

procedure TMDIWallpaper.SetShiftMode(Value: TShiftMode);
begin
  if FShiftMode <> Value then
  begin
    FShiftMode := Value;
    if FMode = wpTile then ClientInvalidate;
  end;
end;

procedure TMDIWallpaper.SetTileWidth(Value: Integer);
begin
  if AutoSizeTile then Exit;
  if Value < Picture.Width then Value := Picture.Width;
  if Value <> FTileWidth then
  begin
    FTileWidth := Value;
    if Mode = wpTile then PictureChanged(Self);
  end;
end;

procedure TMDIWallpaper.SetTileHeight(Value: Integer);
begin
  if AutoSizeTile then Exit;
  if Value < Picture.Height then Value := Picture.Height;
  if Value <> FTileHeight then
  begin
    FTileHeight := Value;
    if Mode = wpTile then PictureChanged(Self);
  end;
end;

procedure TMDIWallpaper.SetTransparent(Value: Boolean);
begin
  if Value <> FTransparent then
  begin
    FTransparent := Value;
    PictureChanged(Self);
  end;
end;

procedure TMDIWallpaper.SetTransparentColor(Value: TColor);
begin
  if Value <> FTransparentColor then
  begin
    if Value = clDefault then
      FTransparentMode := tmAuto else
      FTransparentMode := tmFixed;
    FTransparentColor := Value;
    if Transparent then UpdateWorkingBmp;
  end;
end;

procedure TMDIWallpaper.SetTransparentMode(Value: TTransparentMode);
begin
  if Value <> FTransparentMode then
  begin
    if Value = tmAuto then
      SetTransparentColor(clDefault) else
      SetTransparentColor(GetTransparentColor);
  end;
end;

procedure TMDIWallpaper.SetZigZag(Value: Boolean);
begin
  if Value <> FZigZag then
  begin
    FZigZag := Value;
    if FMode = wpTile then ClientInvalidate;
  end;
end;

function TMDIWallpaper.TransparentColorStored: Boolean;
begin
  Result := FTransparentMode = tmFixed;
end;

procedure TMDIWallpaper.UpdateHOffset;
var
  W: Integer;
begin
  FHOffset := FHorzOffset;
  W := FTileWidth;
  if W = 0 then Exit;
  if FHOffset > 0 then FHOffset := (FHOffset mod W)-W else
  if FHOffset < 0 then Dec(FHOffset, (FHOffset div W)*W);
end;

procedure TMDIWallpaper.UpdateVOffset;
var
  H: Integer;
begin
  FVOffset := FVertOffset;
  H := FTileHeight;
  if H = 0 then Exit;
  if FVOffset > 0 then FVOffset := (FVOffset mod H)-H else
  if FVOffset < 0 then Dec(FVOffset, (FVOffset div H)*H);
end;

{
  TMDIWallpaper.UpdateWorkingBmp
  Transparency: all except TJPEGImage
  GrayMapping: all except TIcon, TMetafile
}
procedure TMDIWallpaper.UpdateWorkingBmp;
var
  X, Y: Integer;
  IsBitmap: Boolean;

  procedure WorkingBmpNeeded;
  begin
    if FWorkingBmp = nil then FWorkingBmp := TBitmap.Create;
    if FMode = wpTile then
    begin
      FWorkingBmp.Width := FTileWidth;
      FWorkingBmp.Height := FTileHeight;
    end else
    begin
      FWorkingBmp.Width := FPicture.Graphic.Width;
      FWorkingBmp.Height := FPicture.Graphic.Height;
    end;
  end;

  procedure DrawGraphic(Graphic: TGraphic);
  begin
    with FWorkingBmp.Canvas do
    begin
      Brush.Color := Client.Color;
      FillRect(Rect(0, 0, FTileWidth, FTileHeight));
      Draw(X, Y, Graphic);
    end;
  end;

  function CreateTransparentBmp(Graphic: TGraphic): TBitmap;
  var
    W, H: Integer;
  begin
    Result := TBitmap.Create;
    if IsBitmap then
      Result.Assign(Graphic)
    else
    begin
      W := Graphic.Width;
      H := Graphic.Height;
      Result.Width := W;
      Result.Height := H;
      with Result.Canvas do
      begin
        Brush.Color := TransparentColor;
        FillRect(Rect(0, 0, W, H));
        Draw(0, 0, Graphic);
      end;
    end;
  end;
{$IFNDEF VER90}
  procedure MapGrays(Bmp: TBitmap);
  var
    Grays: PColorGradation;
    i: Integer;
  begin
    New(Grays);
    try
      for i := Low(Grays^) to High(Grays^) do
        Grays[i] := RGB(i, i, i);
      with Bmp do
      begin
        if ((FPicture.Graphic is TBitmap)
          and (TBitmap(FPicture.Graphic).PixelFormat in [pf1bit..pf8bit]))
        {$IFDEF RECOGNIZE_GIF}
        or (FPicture.Graphic is TGIFImage)
        {$ENDIF}
        then Assign(FPicture.Graphic)
        else
        begin
          PixelFormat := pf8bit;
          Width := FPicture.Width;
          Height := FPicture.Height;
          SetBitmapColors(Bmp, Grays^, 0);
          Canvas.Draw(0, 0, FPicture.Graphic);
        end;
        Handle := CreateMappedBmp(Handle, Grays^, FSysColorGradation^);
      end;
    finally
      Dispose(Grays);
    end;
  end;
{$ENDIF}
var
  Bmp: TBitmap;
{$IFNDEF VER90}
  MaskBmp: TBitmap;
  {$IFNDEF NO_JPEG}
  IsJPEG: Boolean;
  {$ENDIF}
  GrayscaleState: Boolean;
{$ENDIF}
  IsTransparent: Boolean;
  IsTranspGraphic: Boolean;
  IsIcon: Boolean;
  TileTailored: Boolean;
begin
  if FInUpdWorkingBmp then Exit;
  with FPicture do
  if Graphic <> nil then
  try
    FInUpdWorkingBmp := True;
    TileTailored := False;
    X := 0;
    Y := 0;
    if FMode = wpTile then
    begin
      X := FTileWidth-Graphic.Width;
      Y := FTileHeight-Graphic.Height;
      TileTailored := (X <> 0) or (Y <> 0);
      X := X div 2;
      Y := Y div 2;
    end;
    IsBitmap := (Graphic is TBitmap)
                {$IFDEF RECOGNIZE_GIF}
                or (Graphic is TGIFImage) // GIF goes as bitmap here
                {$ENDIF};
    IsIcon := Graphic is TIcon;
    IsTranspGraphic := IsIcon or (Graphic is TMetafile);
    // if Graphic is transparent
{$IFDEF VER90}
    IsTransparent := Transparent or IsTranspGraphic;
    if IsTransparent or TileTailored then
{$ELSE}
    {$IFDEF NO_JPEG}
    IsTransparent := Transparent or IsTranspGraphic;
    {$ELSE}
    IsJPEG := Graphic is TJPEGImage;
    IsTransparent := (Transparent and not IsJPEG) or IsTranspGraphic;
    {$ENDIF}
    if IsTransparent or FGrayMapped or TileTailored then
{$ENDIF}
    begin
      WorkingBmpNeeded;
      if IsTranspGraphic then
      with FWorkingBmp.Canvas do
      begin
        Brush.Color := Client.Color;
        FillRect(Rect(0, 0, FTileWidth, FTileHeight));
        Draw(X, Y, Graphic);
      end else
      if IsTransparent then // and not IsTranspGraphic
      begin
        Bmp := CreateTransparentBmp(Graphic);
        try
          with TImageList.CreateSize(Graphic.Width, Graphic.Height) do
          try
            {$IFDEF VER90}
            AddMasked(Bmp, GetTransparentColor);
            {$ELSE}
            if FGrayMapped then
            begin
              MaskBmp := TBitmap.Create;
              with MaskBmp do
              try
                Assign(Bmp);
                Mask(GetTransparentColor);
                MapGrays(Bmp);
                Add(Bmp, MaskBmp);
              finally
                Free;
              end;
            end
            else
              AddMasked(Bmp, GetTransparentColor);
            FWorkingBmp.HandleType := bmDDB; // otherwise eventually background color won't appear correctly
            {$ENDIF}
            with FWorkingBmp.Canvas do
            begin
              Brush.Color := Client.Color;
              FillRect(Rect(0, 0, FTileWidth, FTileHeight));
            end;
            BkColor := ColorToRGB(Client.Color);
            Draw(FWorkingBmp.Canvas, X, Y, 0);
          finally
            Free;
          end
        finally
          Bmp.Free;
        end
      end
      {$IFNDEF VER90}
      else if GrayMapped then// and not Transparent
      begin
        Bmp := TBitmap.Create;
        try
          {$IFNDEF NO_JPEG}
          if IsJPEG then
          with TJPEGImage(Graphic) do
          begin
            GrayscaleState := Grayscale;
            try
              Grayscale := True;
              Bmp.Assign(Graphic);
            finally
              Grayscale := GrayscaleState;
            end;
          end;
          {$ENDIF}
          MapGrays(Bmp);
          DrawGraphic(Bmp);
        finally
          Bmp.Free;
        end
      end
      {$ENDIF}
      else // if TileTailored
        DrawGraphic(Picture.Graphic);
      ClientInvalidate;
      Exit;
    end;
  finally
    FInUpdWorkingBmp := False;
  end;
  FWorkingBmp.Free;
  FWorkingBmp := nil;
  ClientInvalidate;
end;

const
{$IFDEF Bilch}
  PalettePage = 'Homebrew';
{$ELSE}
  PalettePage = '3rdParty';
{$ENDIF}
procedure Register;
begin
  RegisterComponents(PalettePage, [TMDIWallpaper]);
  RegisterComponentEditor(TMDIWallpaper, TGraphicEditor);
end;

{$IFNDEF VER90}
initialization
finalization
  if Assigned(FSysColorGradation) then
    Dispose(FSysColorGradation);
{$ENDIF}
end.

